; PASMULT.SC

DYNARRAY benchmark[]

PROC timeon(tag.a)

  benchmark[tag.a] = ticks()

ENDPROC ;


PROC timeoff(tag.a)

  benchmark[tag.a] = (ticks()-benchmark[tag.a])/1000

ENDPROC ;


PROC stuff(matn,NROW,NCOL)       ; fill out matrices
PRIVATE i, j;

MENU {Ask} SELECT "interior" 
     TYPEIN("FAST DELETE")
     [matname] = "~matn"
     DO_IT!
     CLEARALL

  EDIT "interior"
  COPYTOARRAY t

  FOR i FROM 1 TO nrow
    FOR j FROM 1 TO ncol
      t["matname"] = matn
      t["row"]     = i
      t["col"]     = j
      IF rand()>0.0 THEN
        t["value"]   = rand()
      ELSE
        t["value"]   = 0
      ENDIF
      INS
      COPYFROMARRAY t
    ENDFOR
    MESSAGE "Row inserted ", STRVAL(i)
  ENDFOR

  DO_IT!
  CLEARALL

ENDPROC;

PROC qrymult(m1,m2,m3)

  TIMEON("QBEMULT")
  MENU {Ask} SELECT "INTERIOR"   ; CHECK # cols
    MOVETO [matname] CHECK [] = m1+" or "+m2
    [row]     = "calc max as rows"
    [col]     = "calc max as cols"
    DO_IT!
    CLEARALL

  MENU {Ask} SELECT "answer"
    MOVETO [matname] CHECK [] = m1
    [cols]    = "_n"
    DOWN
    MOVETO [matname] CHECK [] = m2
    [rows]    = "_n"
    DO_IT!

  IF NRECORDS("answer") <> 1 THEN
    MESSAGE "Incompatible Matrices, cannot multiply"
    SLEEP 1000
    RETURN False
  ENDIF

  ; echo normal

  MENU {Ask} SELECT "interior"
    [#]          = "FAST DELETE"
    [matname] = m3
    DO_IT!
  CLEARALL

  CREATE "MULT"
         "MATNAME" :"A10*",
         "ROW" :"S*",
         "COL1":"S*",
         "COL2":"S*",
         "VALUE":"N"

  MENU {Ask} SELECT "MULT"
    [] = "FAST INSERT"
    [matname] = M3
    [Row] = "_I"
    [Col1] = "_K"
    [Col2] = "_J"
    [Value]= "_val1*_val2"
  MENU {Ask} SELECT "interior"
    [matname] = m1
    [row]     = "_I"
    [col]     = "_K"
    [value]   = "_val1"
    DOWN
    [matname] = m2
    [row]     = "_K"
    [col]     = "_J"
    [value]   = "_val2"
    DO_IT!

  CLEARALL

  MENU {Ask} SELECT "MULT"
    MOVETO [matname] CHECK
    MOVETO [row] CHECK
    MOVETO [col2] CHECK [] = "as Col"
            [value] = "calc sum as Value"
  DO_IT!
  CLEARALL

  ADD "ANSWER" "INTERIOR"
  TIMEOFF("QBEMULT")

ENDPROC ;


PROC ArrayPointerRead(Matname,Rows,Cols)
PRIVATE Buf.a,Irow,ICol

   VIEW "Interior"
   MOVETO [Matname]
   LOCATE Matname

   IF NOT Retval THEN
     MESSAGE "Specified Matrix not found "+Matname
     SLEEP 1000
     RETURN False
   ENDIF

   VARASSIGN ARRAY MatName SIZE Rows

   FOR IROW From 1 TO Rows
     ROWPTR  = MATNAME+STRVAL(IROW)
     VARASSIGN MATNAME INDEX IROW FROM ROWPTR
     VARASSIGN ARRAY ROWPTR SIZE COLS
   ENDFOR

   WHILE ([Matname] = Matname) AND NOT EOT()
     COPYTOARRAY Buf.a
     ROWPTR  = MATNAME+STRVAL(buf.a["ROW"])
     VARASSIGN ROWPTR INDEX buf.a["COL"] FROM buf.a["VALUE"]
     SKIP
   ENDWHILE

   CLEARALL
   RETURN True

ENDPROC ;


PROC ArrayPointerReadCol(Matname,Rows,Cols)
PRIVATE ICol,IRow,Buf

   VIEW "Interior"
   MOVETO [Matname]
   LOCATE Matname
   IF NOT Retval THEN
     MESSAGE "Specified Matrix not found "+Matname
     SLEEP 1000
     RETURN False
   ENDIF

   VARASSIGN ARRAY MatName SIZE Rows
   FOR ICOL From 1 TO Cols
     COLPTR  = MATNAME+STRVAL(ICOL)
     VARASSIGN MATNAME INDEX ICOL FROM COLPTR
     VARASSIGN ARRAY COLPTR SIZE ROWS
   ENDFOR

   WHILE ([Matname] = Matname) AND NOT EOT()
     COPYTOARRAY Buf
     COLPTR  = MATNAME+STRVAL(buf["COL"])
     VARASSIGN COLPTR INDEX buf["ROW"] FROM buf["VALUE"]
     SKIP
   ENDWHILE

   CLEARALL
   RETURN True

ENDPROC ;


PROC ArrayPointerWrite(Matname,Rows,Cols)
PRIVATE I,K,J,Buf

  DYNARRAY BUF[]

  MENU {Ask} SELECT "Interior"
    [#] = "FAST DELETE"
    [Matname] = Matname
    DO_IT!

  EDIT "Interior"
  FOR I FROM 1 TO Rows
    FOR J FROM 1 TO Cols
      Buf["Matname"] = Matname
      Buf["Row"]     = I
      Buf["Col"]     = J
      Buf["Value"]   = VARVAL(VARVAL(MATNAME,I),J)
      INS
      COPYFROMDYNARRAY Buf
    ENDFOR
  ENDFOR

  DO_IT!

  RETURN True

ENDPROC ;


PROC ArrayPointerMult(M1,M2,M3)
PRIVATE i,j,k,m,n,r,mat.rows,mat.cols,
        ROWPTR,M1RWOPTR,M2COLPTR

  DYNARRAY mat.rows[]
  DYNARRAY mat.cols[]
  ; get dimensions

  TIMEON("POINTERMULT")
  MENU {Ask} SELECT "INTERIOR"   ; CHECK # cols
    MOVETO [matname] CHECK [] = m1+" or "+m2
    [row]     = "calc max as rows"
    [col]     = "calc max as cols"
    DO_IT!

  MOVETO "ANSWER"
  SCAN
    mat.rows[[matname]] = [rows]
    mat.cols[[matname]] = [cols]
  ENDSCAN
  CLEARALL

  mat.rows[M3] = mat.rows[M1]
  mat.cols[M3] = mat.cols[M2]

  isok.l = ArrayPointerRead(m1,mat.rows[m1],mat.cols[m1])
  isok.l = isok.l AND
             ArrayPointerReadCol(m2,mat.rows[m2],mat.cols[m2])

  IF NOT Isok.l THEN
    RETURN
  ENDIF

  ; assign matrix M3.
  VARASSIGN ARRAY M3 SIZE mat.Rows[m3]
  FOR IROW From 1 TO mat.Rows[m3]
    ROWPTR  = M3+STRVAL(IROW)
    VARASSIGN M3 INDEX IROW FROM ROWPTR
    VARASSIGN ARRAY ROWPTR SIZE mat.cols[m3]
  ENDFOR

  FOR I FROM 1 TO mat.rows[m1]
    M1ROWPTR = VARVAL(M1,I)
    FOR J FROM 1 TO mat.cols[m2]
      M2COLPTR = VARVAL(M2,J)
      Temp = 0
      FOR K FROM 1 TO Mat.Cols[M1]
        Temp = Temp + VARVAL(M1ROWPTR,K)*VARVAL(M2COLPTR,K)
      ENDFOR
      VARASSIGN VARVAL(M3,I) INDEX J FROM Temp
    ENDFOR
  ENDFOR

  isok.l = ArrayPointerWrite(m3,mat.rows[m3],mat.cols[m3])

  TIMEOFF("POINTERMULT")
  CLEARALL

  RETURN Isok.l

ENDPROC ;

PROC ArraySingleRead(Matname,Rows,Cols)
PRIVATE Buf

  VIEW "Interior"
  MOVETO [Matname]
  LOCATE Matname
  IF NOT Retval THEN
    MESSAGE "Specified Matrix not found "+Matname
    SLEEP 1000
    RETURN False
  ENDIF

  VARASSIGN ARRAY MatName SIZE Rows*Cols
  WHILE ([Matname] = Matname) AND NOT EOT()
    COPYTODYNARRAY Buf
    K = (Buf["Row"]-1)*Cols+buf["Col"]
    VARASSIGN Matname INDEX K FROM buf["VALUE"]
    SKIP
  ENDWHILE

  CLEARALL
  RETURN True

ENDPROC ;


PROC ArraySingleReadCol(Matname,Rows,Cols)
PRIVATE Buf

  VIEW "Interior"
  MOVETO [Matname]
  LOCATE Matname
  IF NOT Retval THEN
    MESSAGE "Specified Matrix not found "+Matname
    SLEEP 1000
    RETURN False
  ENDIF

  VARASSIGN ARRAY MatName SIZE Rows*Cols
  WHILE ([Matname] = Matname) AND NOT EOT()
    COPYTODYNARRAY Buf
    K = (Buf["Col"]-1)*Rows+Buf["Row"]
    VARASSIGN Matname INDEX K FROM buf["VALUE"]
    SKIP
  ENDWHILE

  CLEARALL
  RETURN True

ENDPROC ;


PROC ArraySingleWrite(Matname,Rows,Cols)
PRIVATE I,K,J,Buf

  DYNARRAY BUF[]

  MENU {Ask} SELECT "Interior"
    [#] = "FAST DELETE"
    [Matname] = MAtname
    DO_IT!

  EDIT "Interior"

  FOR I FROM 1 TO Rows
    FOR J FROM 1 TO Cols
      Buf["Matname"] = Matname
      Buf["Row"]     = I
      Buf["Col"]     = J
      Buf["Value"]   = VARVAL(Matname,(I-1)*cols+J)
      INS
      COPYFROMDYNARRAY Buf
    ENDFOR
  ENDFOR

   DO_IT!
   CLEARALL

   RETURN True

ENDPROC ;


PROC ArraySingleMult(M1,M2,M3)
PRIVATE i,j,k,m,n,r,mat.rows,mat.cols,
        ofs1,ofs2,ofs3,Temp

  DYNARRAY mat.rows[]
  DYNARRAY mat.cols[]
  ; get dimensions

  TIMEON("SingleMULT")

  MENU {Ask} SELECT "INTERIOR"   ; CHECK # cols
    MOVETO [matname] CHECK [] = m1+" or "+m2
    [row]     = "calc max as rows"
    [col]     = "calc max as cols"
    DO_IT!

  MOVETO "ANSWER"
  SCAN
    mat.rows[[matname]] = [rows]
    mat.cols[[matname]] = [cols]
  ENDSCAN
  CLEARALL

  isok.l = ArraySingleRead(m1,mat.rows[m1],mat.cols[m1])
  isok.l = isok.l AND
             ArraySingleReadCol(m2,mat.rows[m2],mat.cols[m2])

  IF NOT Isok.l THEN
    RETURN
  ENDIF

  VARASSIGN ARRAY M3 SIZE mat.rows[M1]*mat.cols[M2]

  FOR I FROM 1 TO mat.rows[M1]
    Ofs1 = (I-1)*Mat.Cols[M1]
    FOR J FROM 1 TO mat.cols[M2]
      Ofs2 = (J-1)*Mat.Rows[M2]
      Ofs3 = Ofs1+J
      TEMP = 0.0
      FOR K FROM 1 TO mat.rows[M2]
        Temp = Temp + VARVAL(M1,Ofs1+K)*
                      VARVAL(M2,Ofs2+K)
      ENDFOR
      VARASSIGN M3 INDEX Ofs3 FROM Temp
    ENDFOR
  ENDFOR

  IsOk.l = ArraySingleWrite(M3,Mat.Rows[M1],Mat.Cols[M2])
  CLEARALL

  TimeOff("SingleMULT")

ENDPROC ;ArraySingleMult


PROC DynarrayRead(Matname,Rows,Cols)
PRIVATE isok.l,Tag.a,Buf

  VARASSIGN DYNARRAY Matname
  view "interior"

  MOVETO [Matname] Locate matname
  IF NOT RETVAL THEN
    MESSAGE "Specified Matrix not found "+Matname
    SLEEP 1000
    RETURN False
  ENDIF

  WHILE ([matname] = matname) AND NOT eot()
    copytodynarray buf
    tag.a = STRVAL(buf["row"])+"%"+STRVAL(buf["col"])
    VARASSIGN MATNAME INDEX Tag.a FROM buf["value"]
    SKIP
  ENDWHILE

  CLEARALL
  RETURN True

ENDPROC ; DynarrayRead(matname)


PROC DynarrayWrite(Matname,Rows,Cols)
PRIVATE isok.l,Tag.a,Buf,i.a.

  DYNARRAY Buf[]

  MENU {Ask} SELECT "interior"
    TYPEIN("FAST DELETE")
    [matname] = matname
    DO_IT!
    CLEARALL

  EDIT "interior"
  
  FOR I FROM 1 TO Rows
    I.A = STRVAL(I)
    FOR J FROM 1 TO Cols
      tag.a = I.A+"%"+STRVAL(J)
      buf["matname"] = Matname
      buf["row"]     = i
      buf["col"]     = j
      buf["value"]   = VARVAL(Matname,tag.a)
      INS
      COPYFROMDYNARRAY Buf
    ENDFOR
  ENDFOR

  DO_IT!

  CLEARALL
  RETURN True

ENDPROC ; DynarrayWrite(matname)


PROC DynarrayMult(m1,m2,m3)
PRIVATE i,k,j,mat.rows,mat.cols,
        atag.a,btag.a,ctag.a,result

  DYNARRAY mat.rows[]
  DYNARRAY mat.cols[]
  DYNARRAY result[]
  ; get dimensions

  TIMEON("DynarrayMULT")
  MENU {Ask} SELECT "INTERIOR"   ; CHECK # cols
    MOVETO [matname] CHECK [] = m1+" or "+m2
    [row]     = "calc max as rows"
    [col]     = "calc max as cols"
    DO_IT!

  CLEARALL
  VIEW "ANSWER"

  SCAN
    mat.rows[[matname]] = [rows]
    mat.cols[[matname]] = [cols]
  ENDSCAN
  CLEARALL

  isok.l = DynArrayRead(m1,mat.rows[m1],mat.cols[m1])
  isok.l = isok.l AND
             DynArrayRead(m2,mat.rows[m2],mat.cols[m2])
  IF NOT Isok.l THEN
    RETURN
  ENDIF

  VARASSIGN DYNARRAY M3

  FOR I FROM 1 TO mat.rows[M1]
    I.a = STRVAL(I)
    FOR J FROM 1 TO mat.cols[M2]
      J.A = STRVAL(J)
      M3Tag.a = I.A+"%"+J.A
      Temp = 0
      FOR K FROM 1 TO mat.rows[M2]
        K.A = STRVAL(K)
        Temp = Temp + VARVAL(M1,I.A+"%"+K.A)*
                      VARVAL(M2,K.A+"%"+J.A)
      ENDFOR
      VARASSIGN M3 INDEX M3Tag.a FROM Temp
    ENDFOR
  ENDFOR

isok.l = DynArrayWrite(m3,mat.rows[m1],mat.cols[m1])
Timeoff("DynarrayMult")

ENDPROC ; DynarrayMult()


PROC ExternalExport(Matname,Rows,Cols)
PRIVATE tmpint.t

  tmpint.t = LTRIM(SUBSTR(matname,1,8))
  CREATE tmpint.t like "INTERIOR"

  MENU {Ask} SELECT "interior"
    [matname] = "_n,"+matname
    [row]     = "_r"
    [col]     = "_c"
    [Value]   = "_v"
    
  MENU {Ask} SELECT tmpint.t
    TYPEIN("FAST insert")
    [matname] = "_n"
    [row]     = "_r"
    [col]     = "_c"
    [Value]   = "_v"
    
  DO_IT!
  CLEARALL
    
  isok.l = (rows = cmax(tmpint.t,"ROW")) AND
           (cols = cmax(tmpint.t,"COL"))

  COPYREPORT "interior" "R" tmpint.t "R"
  run norefresh "del "+tmpint.t+".txt > nul"

  MENU {Report} {Output} SELECT tmpint.t SELECT "R"
       {File} SELECT tmpint.t+".txt"
  CLEARALL

  RETURN True

ENDPROC ; ExternalExport


PROC ExternalImport(matname,rows,cols)

  ;ECHO SLOW
  IF isfile("IMPORTED.DB") THEN
    DELETE "imported" 
  ENDIF

  IF isfile("pasmult.out") THEN
    MESSAGE "Importing data FROM external routine"
    MENU {Tools} {ExportImport} {Import} {ASCII} {Delimited} 
      SELECT "pasmult.out" SELECT "imported"

    MESSAGE "Inserting in main files"
    MENU {Ask} SELECT "interior"
      TYPEIN("FAST DELETE")
      [matname] = matname
    DO_IT!
    CLEARALL

    MENU {Ask} SELECT "interior"
      TYPEIN("FAST insert")
      [matname] = matname
      [row]     = "_r"
      [col]     = "_c"
      [Value]   = "_v"
    
    MENU {Ask} SELECT "imported"
      [field-1]     = "_r"
      [field-2]     = "_c"
      [field-3]     = "_v"
      DO_IT!
      CLEARALL
      RETURN True
  ELSE
    RETURN False
  ENDIF

ENDPROC ; Externalimport


PROC ExternalMult(M1,M2,M3)
PRIVATE isok.l

  DYNARRAY mat.rows[]
  DYNARRAY mat.cols[]

  TIMEON("ExternalMULT")

  MENU {Ask} SELECT "INTERIOR"   ; Check # cols
    MOVETO [matname] CHECK [] = m1+" or "+m2
    [row] = "calc max as rows"
    [col] = "calc max as cols"
    DO_IT!

  CLEARALL
  VIEW "ANSWER"

  SCAN
    mat.rows[[matname]] = [rows]
    mat.cols[[matname]] = [cols]
  ENDSCAN
  CLEARALL

  isok.l = ExternalExport(M1,mat.rows[M1],mat.cols[M1])
  isok.l = ExternalExport(M2,mat.rows[M2],mat.cols[M2])

  RUN BIG norefresh "PASMULT " + M1 + ".TXT " + M2 +
                    ".TXT " + M3 + ".TXT "
                    +STRVAL(mat.rows[M1])+" "
                    +STRVAL(mat.cols[M1])+" "
                    +STRVAL(mat.cols[M2])


  isok.l = ExternalImport(M3,mat.rows[M1],mat.cols[M2])
  timeoff("externalmult")

ENDPROC   ; Externalmult


; *************************************************
; *************************************************
; *************************************************
; *************************************************


DYNARRAY results[]

FOR Turn FROM 1 To 10
  DYNARRAY Benchmark[]

  N = 5*Turn

  STUFF("A",N,N)
  STUFF("B",N,N)

  MESSAGE "QRYMULT(\"A\",\"B\",\"C1\")"
  QRYMULT("A","B","C1")
  RELEASE VARS A,B,C1

  MESSAGE "ArraySingleMult(\"A\",\"B\",\"C2\")"
  ArraySingleMult("A","B","C2")
  RELEASE VARS A,B,C2

  MESSAGE "ArrayPointerMult(\"A\",\"B\",\"C3\")"
  ArrayPointerMult("A","B","C3")
  RELEASE VARS A,B,C3

  MESSAGE "DynarrayMult(\"A\",\"B\",\"C4\")"
  DynArrayMult("A","B","C4")
  release vars a,b,c4

  MESSAGE "ExternalMult(\"A\",\"B\",\"C5\")"

  ExternalMult("A","B","C5")

  FOREACH Tag.a in Benchmark
    Results[Tag.a + "." + 
            SUBSTR("00"+Strval(n),
                   3-LEN(STRVAL(n)),2)] = Benchmark[tag.a]
  ENDFOREACH

  SAVEVARS Results

ENDFOR

